from decimal import Decimal
from functools import partial
from pathlib import Path

from hypothesis import given
from hypothesis import strategies as st

from jubeatools import song
from jubeatools.formats import Format
from jubeatools.formats.yubiosi.dump import dump_bgm_start
from jubeatools.formats.yubiosi.load import load_preview_start
from jubeatools.testutils import strategies as jbst
from jubeatools.testutils.test_patterns import dump_and_load_then_compare


@st.composite
def metadata_1_0(draw: st.DrawFn) -> song.Metadata:
    return song.Metadata(title=draw(jbst.shift_jis_safe_text()))


just_16ths = partial(jbst.beat_time, denominator_strat=st.just(4))


@st.composite
def yubiosi_1_0_song(draw: st.DrawFn) -> song.Song:
    """Yubiosi 1.0 files are very limited :
    - one chart
    - no dif info
    - only 16th notes
    - only tap notes
    - no bpm changes
    - only a title
    """
    return draw(
        jbst.song(
            diffs_strat=st.just(set([song.Difficulty.EXTREME.value])),
            common_timing_strat=st.none(),
            common_hakus_strat=st.none(),
            chart_strat=jbst.chart(
                level_strat=st.just(Decimal(0)),
                timing_strat=jbst.timing_info(with_bpm_changes=False),
                notes_strat=jbst.notes(
                    note_strat=jbst.tap_note(time_strat=just_16ths()),
                    beat_time_strat=just_16ths(),
                ),
            ),
            metadata_strat=metadata_1_0(),
        )
    )


@given(yubiosi_1_0_song())
def test_that_full_chart_roundtrips_1_0(s: song.Song) -> None:
    dump_and_load_then_compare(
        Format.YUBIOSI_1_0,
        s,
        bytes_decoder=lambda b: b.decode("shift-jis-2004"),
    )


@st.composite
def metadata_1_5(draw: st.DrawFn) -> song.Metadata:
    return song.Metadata(
        title=draw(jbst.shift_jis_safe_text()),
        audio=Path(draw(jbst.shift_jis_safe_text())),
    )


@st.composite
def safe_bpms(draw: st.DrawFn) -> Decimal:
    """Draw from a selection of bpms with a finite decimal expansion and whole
    number of ms per 1/4 note"""
    val = draw(
        st.sampled_from(
            [
                "75",
                "78.125",
                "93.75",
                "100",
                "117.1875",
                "120",
                "125",
                "150",
                "156.25",
                "187.5",
                "200",
                "234.375",
                "250",
            ]
        )
    )
    return Decimal(val)


@st.composite
def yubiosi_1_5_song(draw: st.DrawFn) -> song.Song:
    """Yubiosi 1.5 files are a bit less restricted :
    - integer level in the 1-10 range
    - bpm changes allowed
    - audio path
    """
    return draw(
        jbst.song(
            diffs_strat=st.just(set([song.Difficulty.EXTREME.value])),
            common_timing_strat=st.none(),
            common_hakus_strat=st.none(),
            chart_strat=jbst.chart(
                level_strat=st.integers(min_value=1, max_value=10),
                timing_strat=jbst.timing_info(
                    with_bpm_changes=True,
                    bpm_strat=safe_bpms(),
                    beat_zero_offset_strat=st.decimals(
                        min_value=0, max_value=20, places=2
                    ),
                    time_strat=just_16ths(min_section=1),
                ),
                notes_strat=jbst.notes(
                    note_strat=jbst.tap_note(time_strat=just_16ths()),
                    beat_time_strat=just_16ths(),
                ),
            ),
            metadata_strat=metadata_1_5(),
        )
    )


@given(yubiosi_1_5_song())
def test_that_full_chart_roundtrips_1_5(s: song.Song) -> None:
    dump_and_load_then_compare(
        Format.YUBIOSI_1_5, s, bytes_decoder=lambda b: b.decode("shift-jis-2004")
    )


@st.composite
def yubiosi_diffs(draw: st.DrawFn) -> set[str]:
    normal_diffs = draw(
        st.sets(
            st.sampled_from(list(d.value for d in song.Difficulty)),
            min_size=0,
            max_size=3,
        )
    )
    if not normal_diffs:
        min_edits = 1
    else:
        min_edits = 0

    edits = draw(st.integers(min_value=min_edits, max_value=7))
    return normal_diffs | {f"EDIT-{i + 1}" for i in range(edits)}


seconds_time = partial(
    st.decimals,
    min_value=0,
    max_value=10**25,
    allow_nan=False,
    allow_infinity=False,
    places=2,
)


@st.composite
def metadata_2_0(draw: st.DrawFn) -> song.Metadata:
    return song.Metadata(
        title=draw(jbst.metadata_text()),
        artist=draw(jbst.metadata_text()),
        audio=Path(draw(jbst.metadata_path())),
        cover=Path(draw(jbst.metadata_path())),
        preview=song.Preview(
            start=draw(seconds_time()),
            length=Decimal(15),
        ),
        preview_file=None,
    )


@st.composite
def yubiosi_2_0_song(draw: st.DrawFn) -> song.Song:
    """Yubiosi 2.0 files have more stuff :
    - utf-16 support
    - artist tag
    - jacket tag
    - audio tag
    - preview start tag
    - sets of charts, with the usual 3 difficulties + 7 optional edits
    - integer levels extented to the 1-99 range
    """
    return draw(
        jbst.song(
            diffs_strat=yubiosi_diffs(),
            common_timing_strat=jbst.timing_info(
                with_bpm_changes=True,
                bpm_strat=safe_bpms(),
                beat_zero_offset_strat=st.decimals(min_value=0, max_value=20, places=2),
                time_strat=just_16ths(min_section=1),
            ),
            common_hakus_strat=st.none(),
            chart_strat=jbst.chart(
                level_strat=st.integers(min_value=1, max_value=99),
                timing_strat=st.none(),
                notes_strat=jbst.notes(
                    note_strat=jbst.tap_note(time_strat=just_16ths()),
                    beat_time_strat=just_16ths(),
                ),
            ),
            metadata_strat=metadata_2_0(),
        )
    )


@given(seconds_time())
def test_that_bgm_start_roundtrips(expected: Decimal) -> None:
    text = dump_bgm_start(expected)
    actual = load_preview_start(text)
    assert actual == expected


@given(yubiosi_2_0_song())
def test_that_full_chart_roundtrips_2_0(s: song.Song) -> None:
    dump_and_load_then_compare(
        Format.YUBIOSI_2_0, s, bytes_decoder=lambda b: b.decode("utf-16")
    )



"""General utility functions"""

import unicodedata
from collections import defaultdict
from decimal import Decimal
from fractions import Fraction
from functools import reduce
from math import gcd
from typing import (
    Any,
    Callable,
    Dict,
    Hashable,
    Iterable,
    List,
    Optional,
    Protocol,
    TypeVar,
)


def single_lcm(a: int, b: int) -> int:
    """Return lowest common multiple of two numbers"""
    return a * b // gcd(a, b)


def lcm(*args: int) -> int:
    """Return lcm of args."""
    return reduce(single_lcm, args, 1)


class SupportsLessThan(Protocol):
    def __lt__(self, __other: Any) -> bool:
        ...


SupportsLessThanT = TypeVar("SupportsLessThanT", bound=SupportsLessThan)  # noqa: Y001


def clamp(
    value: SupportsLessThanT, min_: SupportsLessThanT, max_: SupportsLessThanT
) -> Any:
    return max(min_, min(max_, value))


def charinfo(c: str) -> str:
    """Return some info on the character"""
    return f"{c!r}  # U+{ord(c):05X} : {unicodedata.name(c)}"


A = TypeVar("A")
B = TypeVar("B")

# Monadic stuff !
def none_or(c: Callable[[A], B], e: Optional[A]) -> Optional[B]:
    if e is None:
        return None
    else:
        return c(e)


def fraction_to_decimal(frac: Fraction) -> Decimal:
    "Thanks stackoverflow ! https://stackoverflow.com/a/40468867/10768117"
    return frac.numerator / Decimal(frac.denominator)


K = TypeVar("K", bound=Hashable)
V = TypeVar("V")


def group_by(elements: Iterable[V], key: Callable[[V], K]) -> Dict[K, List[V]]:
    res = defaultdict(list)
    for e in elements:
        res[key(e)].append(e)

    return res


V2 = TypeVar("V2", bound=Hashable)


def reverse_dict(d: dict[K, V2]) -> dict[V2, K]:
    return {v: k for k, v in d.items()}


def mixed_number_format(f: Fraction) -> str:
    """Formats fractions the following way :

    >>> mixed_number_format(Fraction(4,2))
    '2'
    >>> mixed_number_format(Fraction(-12,3))
    '-4'
    >>> mixed_number_format(Fraction(1,3))
    '1/3'
    >>> mixed_number_format(Fraction(-1,3))
    '-1/3'
    >>> mixed_number_format(Fraction(5,3))
    '1 + 2/3'
    >>> mixed_number_format(Fraction(-7,3))
    '-(2 + 1/3)'
    """
    abs_value = abs(f)
    integer_part = int(abs_value)
    fractional_part = abs_value % 1
    addition_elements = []
    if integer_part != 0:
        addition_elements.append(str(integer_part))
    if fractional_part != 0:
        addition_elements.append(
            f"{fractional_part.numerator}/{fractional_part.denominator}"
        )
    if not addition_elements:
        return "0"

    addition = " + ".join(addition_elements)
    if f >= 0:
        return addition
    else:
        if len(addition_elements) > 1:
            return f"-({addition})"
        else:
            return f"-{addition}"



"""
yubiosi (指押) is a now-defunct jubeat simulator for Android

It notably also supports 3x3 charts although jubeatools only supports
converting to and from 4x4 charts

The three versions of the format are documented on a dedicated wiki :

https://w.atwiki.jp/yubiosi2/
"""

from .dump import dump_yubiosi_1_0, dump_yubiosi_1_5, dump_yubiosi_2_0
from .load import load_yubiosi_1_0, load_yubiosi_1_5, load_yubiosi_2_0



import json
import re
from functools import wraps
from io import StringIO
from pathlib import Path
from typing import Any, Callable, Type

from .format_names import Format


def guess_format(path: Path) -> Format:
    """Try to guess the format of the given file, raise an exception if the
    format is unknown"""
    if path.is_dir():
        raise ValueError("Can't guess chart format for a folder")

    contents = path.read_bytes()

    return guess_file_format(contents)


def guess_file_format(contents: bytes) -> Format:
    try:
        return recognize_json_formats(contents)
    except (json.JSONDecodeError, UnicodeError, ValueError):
        pass

    try:
        return recognize_jubeat_analyser_format(contents)
    except (UnicodeError, ValueError):
        pass

    if looks_like_eve(contents):
        return Format.EVE

    if looks_like_jbsq(contents):
        return Format.JBSQ

    if looks_like_yubiosi_1_0(contents):
        return Format.YUBIOSI_1_0

    if looks_like_yubiosi_1_5(contents):
        return Format.YUBIOSI_1_5

    if looks_like_yubiosi_2_0(contents):
        return Format.YUBIOSI_2_0

    raise ValueError("Unrecognized file format")


def recognize_json_formats(contents: bytes) -> Format:
    obj = json.loads(contents)

    if not isinstance(obj, dict):
        raise ValueError("Top level value is not an object")

    if obj.keys() & {"metadata", "data", "version"}:
        return recognize_memon_version(obj)
    elif obj.keys() >= {"meta", "time", "note"}:
        return Format.MALODY
    else:
        raise ValueError("Unrecognized file format")


def recognize_memon_version(obj: dict) -> Format:
    try:
        version = obj["version"]
    except KeyError:
        return Format.MEMON_LEGACY

    if version == "0.1.0":
        return Format.MEMON_0_1_0
    elif version == "0.2.0":
        return Format.MEMON_0_2_0
    elif version == "0.3.0":
        return Format.MEMON_0_3_0
    elif version == "1.0.0":
        return Format.MEMON_1_0_0
    else:
        raise ValueError(f"Unsupported memon version : {version}")


JUBEAT_ANALYSER_COMMANDS = {
    "b",
    "m",
    "o",
    "r",
    "t",
    "#lev",
    "#dif",
    "#title",
    "#artist",
}

COMMENT = re.compile(r"//.*")


def _dirty_jba_line_strip(line: str) -> str:
    """This does not deal with '//' in quotes properly,
    thankfully we don't care when looking for an argument-less command"""
    return COMMENT.sub("", line).strip()


def recognize_jubeat_analyser_format(contents: bytes) -> Format:
    text = contents.decode(encoding="shift-jis-2004", errors="surrogateescape")
    lines = text.splitlines()

    saw_jubeat_analyser_commands = False
    for raw_line in lines:
        line = _dirty_jba_line_strip(raw_line)
        if line in ("#memo2", "#boogie"):
            return Format.MEMO_2
        elif line == "#memo1":
            return Format.MEMO_1
        elif line == "#memo":
            return Format.MEMO
        elif "=" in line:
            index = line.index("=")
            if line[:index] in JUBEAT_ANALYSER_COMMANDS:
                saw_jubeat_analyser_commands = True

    if saw_jubeat_analyser_commands:
        return Format.MONO_COLUMN
    else:
        raise ValueError("Unrecognized file format")


def false_if_raises(
    *exceptions: Type[Exception],
) -> Callable[[Callable[..., bool]], Callable[..., bool]]:
    def decorator(f: Callable[..., bool]) -> Callable[..., bool]:
        @wraps(f)
        def wrapper(*a: Any, **kw: Any) -> bool:
            try:
                return f(*a, **kw)
            except Exception as e:
                if exceptions and not isinstance(e, exceptions):
                    raise
                else:
                    return False

        return wrapper

    return decorator


@false_if_raises(UnicodeError, StopIteration)
def looks_like_eve(contents: bytes) -> bool:
    f = StringIO(contents.decode("ascii"))
    return looks_like_eve_line(f.readline())


EVE_COMMANDS = {
    "END",
    "MEASURE",
    "HAKU",
    "PLAY",
    "LONG",
    "TEMPO",
}


def looks_like_eve_line(line: str) -> bool:
    columns = line.split(",")
    if len(columns) != 3:
        return False

    raw_tick, raw_command, raw_value = map(str.strip, columns)
    try:
        int(raw_tick)
    except Exception:
        return False

    if raw_command not in EVE_COMMANDS:
        return False

    try:
        int(raw_value)
    except Exception:
        return False

    return True


def looks_like_jbsq(contents: bytes) -> bool:
    magic = contents[:4]
    return magic in (b"IJBQ", b"IJSQ", b"JBSQ")


@false_if_raises(UnicodeError, ValueError)
def looks_like_yubiosi_1_0(contents: bytes) -> bool:
    lines = contents.decode(encoding="shift-jis-2004").splitlines()
    (
        _,  # title
        _,  # save_data_name
        raw_bpm,
        chart_duration_ms,
        raw_offset,
        raw_note_count,
        *raw_times_and_positions,
    ) = lines
    float(raw_bpm)
    int(chart_duration_ms)
    int(raw_offset)
    note_count = int(raw_note_count)
    return len(raw_times_and_positions) == 2 * note_count


YUBIOSI_1_5_TAGS = {
    "TITLE_NAME",
    "DATA_ID",
    "BPM",
    "BPMS",
    "MUSIC_TIME",
    "OFFSET",
    "LEVEL",
    "PANEL_NUM",
    "AUDIO_FILE",
}


@false_if_raises(UnicodeError, ValueError)
def looks_like_yubiosi_1_5(contents: bytes) -> bool:
    lines = contents.decode(encoding="shift-jis-2004").splitlines()
    note_index = lines.index("[Notes]")
    return any(line_has_yubiosi_tag(line) for line in lines[:note_index])


def line_has_yubiosi_tag(line: str) -> bool:
    if line.startswith("#") and ":" in line:
        potential_tag = line[1 : line.index(":")]
        return potential_tag in YUBIOSI_1_5_TAGS
    else:
        return False


@false_if_raises(UnicodeError)
def looks_like_yubiosi_2_0(contents: bytes) -> bool:
    text = contents.decode("utf-16")
    lines = text.splitlines()
    return len(lines) >= 1 and lines[0] == "//Yubiosi 2.0"



from jubeatools import song
from jubeatools.utils import reverse_dict

INDEX_TO_YUBIOSI_1_0 = {n: 2**n for n in range(16)}

YUBIOSI_1_0_TO_INDEX = reverse_dict(INDEX_TO_YUBIOSI_1_0)

INDEX_TO_DIF = {
    1: song.Difficulty.BASIC.value,
    2: song.Difficulty.ADVANCED.value,
    3: song.Difficulty.EXTREME.value,
    4: "EDIT-1",
    5: "EDIT-2",
    6: "EDIT-3",
    7: "EDIT-4",
    8: "EDIT-5",
    9: "EDIT-6",
    10: "EDIT-7",
}



from jubeatools.testutils.resources import load_test_file

from . import data


def test_1_0() -> None:
    """Demo chart from yubiosi's .apk"""
    load_test_file(data, "m1.txt")


def test_1_5() -> None:
    """Transcript of the official chart found on a Czech website :
    http://download.pocitac.com/index.php?dir=JuBeat/Android/"""
    load_test_file(data, "concon (EXT).txt")


def test_2_0() -> None:
    """Custom chart from http://pyagoraa.blog.fc2.com/"""
    song = load_test_file(data, "Oshama Scramble!.ybi")
    assert "EXT" in song.charts



import hypothesis.strategies as st
from hypothesis import given

from .. import guess


@given(st.binary())
def test_that_guess_format_only_raises_the_specific_value_error(
    contents: bytes,
) -> None:
    try:
        guess.guess_file_format(contents)
    except ValueError as e:
        if e.args != ("Unrecognized file format",):
            raise


def test_that_yubiosi_2_0_detection_does_not_raise_exception_for_non_utf16_files() -> None:
    text = "blablabla"
    bytes_ = text.encode("ascii")
    guess.looks_like_yubiosi_2_0(bytes_)



from fractions import Fraction
from math import ceil
from pathlib import Path
from typing import Any, Iterator, List, Sequence, Union
from uuid import uuid4

from jubeatools import song as jbt
from jubeatools.formats.dump_tools import (
    FileNameFormat,
    make_dumper_from_chart_file_dumper,
)
from jubeatools.formats.filetypes import ChartFile
from jubeatools.formats.timemap import TimeMap
from jubeatools.info import __repository_url__, __version__
from jubeatools.utils import clamp, fraction_to_decimal

from .tools import INDEX_TO_YUBIOSI_1_0


def _dump_yubiosi_1_0(song: jbt.Song, **kwargs: dict) -> List[ChartFile]:
    res = []
    for dif, chart, timing, hakus in song.iter_charts():
        if any(isinstance(n, jbt.LongNote) for n in chart.notes):
            raise ValueError("yubiosi 1.0 does not support long notes")

        if len(timing.events) != 1:
            raise ValueError("yubiosi 1.0 does not support BPM changes")

        if any((n.time.denominator not in (1, 2, 4)) for n in chart.notes):
            raise ValueError("yubiosi 1.0 only supports 16th notes")

        timemap = TimeMap.from_timing(timing)
        sorted_notes = sorted(chart.notes, key=lambda n: n.time)
        if sorted_notes:
            last_note = sorted_notes[-1]
            chart_duration = ceil(timemap.seconds_at(last_note.time) * 1000)
        else:
            chart_duration = 0

        header = [
            song.metadata.title or "",
            "",  # save data name
            timing.events[0].BPM,
            chart_duration,  # chart duration in ms
            int(timing.beat_zero_offset * 1000),
            len(chart.notes),
        ]
        times = [int(note.time * 4) for note in sorted_notes]
        positions = [INDEX_TO_YUBIOSI_1_0[note.position.index] for note in sorted_notes]
        chart_lines = header + times + positions
        chart_text = "\n".join(map(str, chart_lines))
        chart_bytes = chart_text.encode("shift-jis-2004")
        res.append(ChartFile(chart_bytes, song, dif, chart))

    return res


dump_yubiosi_1_0 = make_dumper_from_chart_file_dumper(
    internal_dumper=_dump_yubiosi_1_0, file_name_template=Path("{title}.txt")
)


def _dump_yubiosi_1_5(song: jbt.Song, **kwargs: dict) -> List[ChartFile]:
    res = []
    for dif, chart, timing, hakus in song.iter_charts():
        if any(isinstance(n, jbt.LongNote) for n in chart.notes):
            raise ValueError("yubiosi 1.5 does not support long notes")

        if any((n.time.denominator not in (1, 2, 4)) for n in chart.notes):
            raise ValueError("yubiosi 1.5 only supports 16th notes")

        header = list(iter_header_lines_1_5(song.metadata, chart, timing))
        # redundant type check to placate mypy
        notes = [dump_note_1_5(n) for n in chart.notes if isinstance(n, jbt.TapNote)]
        chart_lines = header + ["", "[Notes]"] + notes
        chart_text = "\n".join(map(str, chart_lines))
        chart_bytes = chart_text.encode("shift-jis-2004")
        res.append(ChartFile(chart_bytes, song, dif, chart))

    return res


dump_yubiosi_1_5 = make_dumper_from_chart_file_dumper(
    internal_dumper=_dump_yubiosi_1_5, file_name_template=Path("{title}.txt")
)

TAG_ORDER_1_5 = [
    "TITLE_NAME",
    "DATA_ID",
    "BPMS",
    "MUSIC_TIME",
    "OFFSET",
    "LEVEL",
    "PANEL_NUM",
    "AUDIO_FILE",
]


def iter_header_lines_1_5(
    metadata: jbt.Metadata, chart: jbt.Chart, timing: jbt.Timing
) -> Iterator[str]:
    tags = compute_header_tags_1_5(metadata, chart, timing)
    for key in TAG_ORDER_1_5:
        if key in tags:
            yield dump_tag(key, tags[key])


def dump_tag(key: str, value: str) -> str:
    return f"#{key}:{value}"


def compute_header_tags_1_5(
    metadata: jbt.Metadata, chart: jbt.Chart, timing: jbt.Timing
) -> dict[str, str]:
    timemap = TimeMap.from_timing(timing)
    sorted_notes = sorted(chart.notes, key=lambda n: n.time)
    if sorted_notes:
        last_note = sorted_notes[-1]
        chart_duration = ceil(timemap.seconds_at(last_note.time) * 1000)
    else:
        chart_duration = 0

    tags = {
        "DATA_ID": uuid4().hex,  # just use something unique
        "BPMS": dump_bpms(timemap, Fraction(timing.beat_zero_offset)),
        "MUSIC_TIME": chart_duration,
        "OFFSET": int(timing.beat_zero_offset * 1000),
        "PANEL_NUM": "4x4",
    }

    if metadata.title is not None:
        tags["TITLE_NAME"] = metadata.title

    if chart.level is not None:
        tags["LEVEL"] = clamp(int(chart.level), 1, 10)

    if metadata.audio is not None:
        tags["AUDIO_FILE"] = metadata.audio

    return {k: str(v) for k, v in tags.items()}


def dump_bpms(timemap: TimeMap, offset: Fraction) -> str:
    return ",".join(
        dump_bpm_event(e.BPM, e.seconds - offset) for e in timemap.events_by_seconds
    )


def dump_bpm_event(bpm: Fraction, seconds: Fraction) -> str:
    d_bpm = fraction_to_decimal(bpm)
    ms = int(seconds * 1000)
    return f"{d_bpm}/{ms}"


def dump_note_1_5(note: jbt.TapNote) -> str:
    time = int(note.time * 4)
    position = note.position.index + 1
    return f"{time},{position}"


def dump_yubiosi_2_0(song: jbt.Song, path: Path, **kwargs: dict) -> dict[Path, bytes]:
    info_file_name_format = FileNameFormat(Path("{title}.ybi"), suggestion=path)
    ybi_path = info_file_name_format.available_song_filename(song)
    taken_filenames = {ybi_path}
    edits = sorted(song.charts.keys() - set(d.value for d in jbt.Difficulty))
    if len(edits) > 7:
        raise ValueError(
            f"This song has too many edit difficulties ({len(edits)}), yubiosi "
            "2.0 only supports 7 per song."
        )

    chart_paths = {}
    chart_path_suggestion = path.with_suffix(".ybh") if not path.is_dir() else path
    for dif, chart in song.charts.items():
        chart_file_name_format = FileNameFormat(
            Path("{title}[{difficulty}].ybh"), suggestion=chart_path_suggestion
        )
        chart_filename = chart_file_name_format.available_chart_filename(
            song, dif, chart, already_chosen=taken_filenames
        )
        taken_filenames.add(chart_filename)
        chart_paths[dif] = chart_filename

    files = {ybi_path: dump_yubiosi_2_0_info_file(song, edits, chart_paths)}
    for dif, chart in song.charts.items():
        chart_path = chart_paths[dif]
        files[chart_path] = dump_yubiosi_2_0_chart_file(chart.notes)

    return files


def dump_yubiosi_2_0_info_file(
    song: jbt.Song, edits: list[str], chart_filenames: dict[str, Path]
) -> bytes:
    song.minimize_timings()
    if len(song.charts) > 1 and any(c.timing is not None for c in song.charts.values()):
        raise ValueError("Yubiosi 2.0 does not support per-chart timing")

    if len(song.charts) == 1:
        timing = next(iter(song.charts.values())).timing or song.common_timing
    else:
        timing = song.common_timing

    if timing is None:
        raise ValueError("No timing info")

    lines = [
        "//Yubiosi 2.0",
        f"// Converted using jubeatools {__version__}",
        f"// {__repository_url__}",
    ]
    lines.extend(iter_header_lines_2_0(song, timing))
    diffs_order = list(j.value for j in jbt.Difficulty) + edits
    for i, dif in enumerate(diffs_order, start=1):
        if (chart := song.charts.get(dif)) is not None:
            filename = chart_filenames[dif]
            lines += dump_chart_section(chart, i, filename, timing.beat_zero_offset)

    return "\n".join(lines).encode("utf-16")


TAG_ORDER_2_0 = ["TITLE_NAME", "ARTIST", "BPMS", "AUDIO_FILE", "JACKET", "BGM_START"]


def iter_header_lines_2_0(song: jbt.Song, timing: jbt.Timing) -> Iterator[str]:
    tags = compute_header_tags_2_0(song, timing)
    for key in TAG_ORDER_2_0:
        if key in tags:
            yield dump_tag(key, tags[key])


def compute_header_tags_2_0(song: jbt.Song, timing: jbt.Timing) -> dict[str, str]:
    tags: dict[str, Any] = {}
    if song.metadata.title is not None:
        tags["TITLE_NAME"] = song.metadata.title

    if song.metadata.artist is not None:
        tags["ARTIST"] = song.metadata.artist

    timemap = TimeMap.from_timing(timing)
    tags["BPMS"] = dump_bpms(timemap, Fraction(timing.beat_zero_offset))

    if song.metadata.audio is not None:
        tags["AUDIO_FILE"] = song.metadata.audio

    if song.metadata.cover is not None:
        tags["JACKET"] = song.metadata.cover

    if song.metadata.preview is not None:
        tags["BGM_START"] = dump_bgm_start(song.metadata.preview.start)

    return {k: str(v) for k, v in tags.items()}


def dump_bgm_start(start: jbt.SecondsTime) -> str:
    return str(int(start * 1000))


def dump_chart_section(
    chart: jbt.Chart, index: int, path: Path, offset: jbt.SecondsTime
) -> list[str]:
    return [
        f"[Notes:{index}]",
        dump_tag("FILE", path.name),
        dump_tag("LEVEL", str(clamp(int(chart.level or 0), 1, 99))),
        dump_tag("OFFSET", str(int(offset * 1000))),
    ]


def dump_yubiosi_2_0_chart_file(
    notes: Sequence[Union[jbt.TapNote, jbt.LongNote]]
) -> bytes:
    if any(isinstance(n, jbt.LongNote) for n in notes):
        raise ValueError("yubiosi 2.0 does not support long notes")

    if any((n.time.denominator not in (1, 2, 4)) for n in notes):
        raise ValueError("yubiosi 2.0 only supports 16th notes")

    # redundant type check to placate mypy
    note_lines = [dump_note_1_5(n) for n in notes if isinstance(n, jbt.TapNote)]
    lines = ["//Yubiosi 2.0"] + note_lines
    return "\n".join(lines).encode("utf-16")



import re
import warnings
from collections import defaultdict
from dataclasses import replace
from decimal import Decimal
from fractions import Fraction
from pathlib import Path
from typing import Any, Iterator, NamedTuple, Optional

from jubeatools import song
from jubeatools.formats.load_tools import FolderLoader, make_folder_loader
from jubeatools.formats.timemap import BPMAtSecond, TimeMap
from jubeatools.utils import none_or

from .tools import INDEX_TO_DIF, YUBIOSI_1_0_TO_INDEX


def load_yubiosi_1_0(path: Path, **kwargs: Any) -> song.Song:
    files = load_folder_1_x(path)
    charts = [load_yubiosi_1_0_file(lines) for lines in files.values()]
    return song.Song.from_monochart_instances(*charts)


def load_file_1_x(path: Path) -> list[str]:
    with path.open(encoding="shift-jis-2004") as f:
        return f.read().split("\n")


load_folder_1_x: FolderLoader[list[str]] = make_folder_loader("*.txt", load_file_1_x)


def load_yubiosi_1_0_file(lines: list[str]) -> song.Song:
    (
        title,
        _,  # save data name
        raw_bpm,
        _,  # chart duration in ms
        raw_offset,
        raw_note_count,
        *raw_times_and_positions,
    ) = lines
    note_count = int(raw_note_count)
    if len(raw_times_and_positions) != 2 * note_count:
        raise ValueError(
            "Incorrect line count. The yubiosi file header claims this file "
            f"contains {note_count} notes, this imples the file body should "
            f"have twice as many lines ({note_count * 2}) but found {len(raw_times_and_positions)}"
        )

    bpm = Decimal(raw_bpm)
    offset = song.SecondsTime(raw_offset) / 1000
    timing = song.Timing(
        events=[song.BPMEvent(time=song.BeatsTime(0), BPM=bpm)], beat_zero_offset=offset
    )
    raw_times = raw_times_and_positions[:note_count]
    raw_positions = raw_times_and_positions[note_count:]
    times = [int(t) for t in raw_times]
    positions = [int(p) for p in raw_positions]
    notes = [load_1_0_note(t, p) for t, p in zip(times, positions)]
    chart = song.Chart(level=Decimal(0), timing=timing, notes=notes)
    return song.Song(
        metadata=song.Metadata(title=title),
        charts={song.Difficulty.EXTREME.value: chart},
    )


def load_1_0_note(time: int, position: int) -> song.TapNote:
    time_in_beats = song.BeatsTime(time, 4)
    index = YUBIOSI_1_0_TO_INDEX[position]
    return song.TapNote(
        time=time_in_beats, position=song.NotePosition.from_index(index)
    )


def load_yubiosi_1_5(path: Path, **kwargs: Any) -> song.Song:
    files = load_folder_1_x(path)
    charts = [load_yubiosi_1_5_file(lines) for lines in files.values()]
    return song.Song.from_monochart_instances(*charts)


def load_yubiosi_1_5_file(lines: list[str]) -> song.Song:
    try:
        notes_marker = lines.index("[Notes]")
    except ValueError:
        raise ValueError("No '[Notes]' line found in the file") from None

    tag_lines = lines[:notes_marker]
    note_lines = lines[notes_marker + 1 :]

    tags = dict(iter_tags(tag_lines))
    panel_num = tags.get("PANEL_NUM")
    if panel_num is not None and panel_num != "4x4":
        raise ValueError(f"Unsupported PANEL_NUM : {panel_num}")

    title = tags.get("TITLE_NAME")
    timing = load_timing_data_1_5(tags)
    notes = [load_1_5_note(line) for line in note_lines if line.strip()]
    level = Decimal(tags.get("LEVEL", 0))
    audio = tags.get("AUDIO_FILE")

    return song.Song(
        metadata=song.Metadata(title=title, audio=none_or(Path, audio)),
        charts={
            song.Difficulty.EXTREME.value: song.Chart(
                level=level, timing=timing, notes=notes
            )
        },
    )


TAG_REGEX = re.compile(r"#(?P<key>\w+?):(?P<value>.*)")


def iter_tags(tag_lines: list[str]) -> Iterator[tuple[str, str]]:
    for line in tag_lines:
        if match := TAG_REGEX.fullmatch(line):
            yield match["key"], match["value"]


def load_timing_data_1_5(tags: dict[str, str]) -> song.Timing:
    offset = load_offset(tags)
    events = [replace(e, seconds=e.seconds + offset) for e in iter_bpm_events(tags)]
    return TimeMap.from_seconds(events).convert_to_timing_info()


def load_offset(tags: dict[str, str]) -> Fraction:
    offset_ms = int(tags.get("OFFSET", 0))
    return Fraction(offset_ms, 1000)


def iter_bpm_events(tags: dict[str, str]) -> Iterator[BPMAtSecond]:
    if (raw_bpm_changes := tags.get("BPMS")) is not None:
        yield from iter_bpms_tag(raw_bpm_changes)
    elif (raw_bpm := tags.get("BPM")) is not None:
        yield BPMAtSecond(Fraction(0), Fraction(Decimal(raw_bpm)))
    else:
        raise ValueError("The file is missing either a #BPM or a #BPMS tag")


def iter_bpms_tag(tag_value: str) -> Iterator[BPMAtSecond]:
    for event in tag_value.split(","):
        bpm, ms = event.split("/")
        yield BPMAtSecond(seconds=Fraction(int(ms), 1000), BPM=Fraction(Decimal(bpm)))


def load_1_5_note(line: str) -> song.TapNote:
    raw_time, raw_position = line.split(",")
    return song.TapNote(
        time=song.BeatsTime(int(raw_time), 4),
        position=song.NotePosition.from_index(int(raw_position) - 1),
    )


def load_yubiosi_2_0(path: Path, **kwargs: Any) -> song.Song:
    ybi = find_ybi_file(path)
    header, charts_headers = load_yubiosi_2_0_ybi_file(ybi)
    metadata = load_yubiosi_2_0_metadata(header)
    bpms = list(iter_bpm_events(header))
    ignored_indicies = charts_headers.keys() - INDEX_TO_DIF.keys()
    if ignored_indicies:
        warnings.warn(
            "This file contains [Notes:_] sections with invalid difficulty "
            f"indicies : {ignored_indicies}. [Notes:_] sections with indicies "
            "outside the 1-10 range are ignored."
        )
    charts = {}
    for dif_index in charts_headers.keys() & INDEX_TO_DIF.keys():
        tags = charts_headers[dif_index]
        dif = INDEX_TO_DIF[dif_index]
        chart = load_yubiosi_2_0_chart(tags, bpms, ybi)
        charts[dif] = chart

    return song.Song(metadata=metadata, charts=charts)


def find_ybi_file(path: Path) -> Path:
    if path.is_file():
        if path.suffix != ".ybi":
            warnings.warn(
                "The input file does not have the .ybi extension. Make sure "
                "you gave the path to the metadata file and not the path to "
                "one of the .ybh charts by mistake."
            )
        return path
    elif path.is_dir():
        ybis = list(path.glob("*.ybi"))
        ybis_count = len(ybis)
        if ybis_count > 1:
            raise ValueError(
                "Found multiple .ybi files in the given folder. The given path "
                "should either be a file, or a folder containing only one file "
                "with .ybi extension"
            )
        elif ybis_count == 0:
            raise ValueError(
                "No .ybi files found in the given folder. The given path "
                "should either be a file, or a folder containing only one file "
                "with .ybi extension"
            )
        else:
            return ybis[0]
    else:
        raise ValueError("The source path is neither a file nor a folder")


def load_yubiosi_2_0_ybi_file(
    ybi: Path,
) -> tuple[dict[str, str], dict[int, dict[str, str]]]:
    lines = ybi.read_text(encoding="utf-16").split("\n")
    if not lines:
        raise ValueError("This file is empty")

    warn_for_missing_version_comment(lines)
    chart_headers: defaultdict[int, dict[str, str]] = defaultdict(dict)
    file_header: dict[str, str] = {}
    for dif_index, key, value in iter_section_and_tags(lines):
        if dif_index is None:
            file_header[key] = value
        else:
            chart_headers[dif_index][key] = value

    return file_header, dict(chart_headers)


def warn_for_missing_version_comment(lines: list[str]) -> None:
    if lines[0] != "//Yubiosi 2.0":
        warnings.warn(
            "Unexpected text on the first line of the .ybi file. "
            "The first line should be '//Yubiosi 2.0'"
        )


class TagWithSection(NamedTuple):
    section: Optional[int]
    key: str
    value: str


SECTION_REGEX = re.compile(r"\[Notes:(\d+)\]")


def iter_section_and_tags(lines: list[str]) -> Iterator[TagWithSection]:
    section = None
    for i, line in enumerate(lines, start=1):
        if line.startswith("//"):
            continue
        elif match := SECTION_REGEX.fullmatch(line):
            section = int(match[1])
        elif match := TAG_REGEX.fullmatch(line):
            yield TagWithSection(
                section=section, key=match["key"], value=match["value"]
            )
        elif line.strip():
            warnings.warn(f"Unexpected text on line {i}")


def load_preview_start(val: str) -> Decimal:
    return Decimal(val) / 1000


def load_yubiosi_2_0_metadata(header: dict[str, str]) -> song.Metadata:
    return song.Metadata(
        title=header.get("TITLE_NAME"),
        artist=header.get("ARTIST"),
        audio=none_or(Path, header.get("AUDIO_FILE")),
        cover=none_or(Path, header.get("JACKET")),
        preview=none_or(
            lambda t: song.Preview(start=load_preview_start(t), length=Decimal(15)),
            header.get("BGM_START"),
        ),
    )


def load_yubiosi_2_0_chart(
    tags: dict[str, str], bpms: list[BPMAtSecond], ybi: Path
) -> song.Chart:
    file = ybi.parent / Path(tags["FILE"])
    lines = file.read_text("utf-16").split("\n")
    warn_for_missing_version_comment(lines)
    offset = load_offset(tags)
    timing = TimeMap.from_seconds(
        [replace(e, seconds=e.seconds + offset) for e in bpms]
    ).convert_to_timing_info()
    level = Decimal(tags.get("LEVEL", 0))
    notes = list(iter_yubiosi_2_0_notes(lines))
    return song.Chart(level=level, timing=timing, notes=notes)


def iter_yubiosi_2_0_notes(lines: list[str]) -> Iterator[song.TapNote]:
    for i, line in enumerate(lines, start=1):
        if line.startswith("//"):
            continue
        elif not line.strip():
            continue
        else:
            try:
                yield load_1_5_note(line)
            except Exception:
                warnings.warn(f"Couldn't parse line {i} as a note ... ignoring it")



from jubeatools.testutils.resources import load_test_file

from . import data


def test_Beat_Juggling_Mix() -> None:
    """The loader used to move the common timing into the song timing because
    of a bug"""
    song = load_test_file(data, "Beat Juggling Mix.memon")
    assert song.common_timing is not None
